from zipfile import ZipFile
from cStringIO import StringIO
from string import punctuation
from _base import BaseExporter
from _csv import CSVExporter


class SasExporter(BaseExporter):
    file_extension = 'zip'
    preferred_formats = ('coded', 'number', 'string')

    num_lg_names = 0

    # informat/format mapping for all datatypes except strings
    sas_informat_map = {
        'number': 'best32.',
        'date': 'MMDDYYw.',
        'boolean': 'best32.',
        'datetime': 'DATETIMEw.d',
        'time': 'TIMEw.d'
    }

    sas_format_map = {
        'number': 'best12.',
        'date': 'MMDDYYw.',
        'boolean': 'best12.',
        'datatime': 'DATETIMEw.d',
        'time': 'TIMEw.d'
    }

    def _format_name(self, name):
        punc = punctuation.replace('_', '')
        name = str(name).translate(None, punc)
        name = name.replace(' ', '_')
        if name[0].isdigit():
            name = '_' + name
        if len(name) < 30:
            return name
        else:
            self.num_lg_names += 1
            sas_name = name[:20]
            sas_name += '_lg_{0}'.format(self.num_lg_names)
            return sas_name

    def _get_formats(self, sas_name, datafield):
        """This method creates the sas format and informat lists for
        every variable.
        """
        # get the informat/format
        if datafield.simple_type == 'string':
            informat = s_format = '${0}.'.format(datafield.field.max_length)
        else:
            s_format = self.sas_format_map[datafield.simple_type]
            informat = self.sas_informat_map[datafield.simple_type]

        sas_informat = '\tinformat {0:<10}{1:>10};\n'.format(sas_name, informat)
        sas_format = '\tformat {0:<10}{1:>10};\n'.format(sas_name, s_format)
        return sas_format, sas_informat

    def _code_values(self, name, field):
        """If field can be coded return the value dictionary
        and the format name for the dictionary
        """
        value_format = '\tformat {0} {0}_f.;\n'.format(name)
        value = '\tvalue {0}_f '.format(name)
        values_len = len(field.codes)

        for i, (val, code) in enumerate(field.codes):
            value += '{0}="{1}" '.format(code, val)
            if (i != values_len - 1 and (i % 2) == 1 and i != 0):
                value += '\n\t\t'
        value += ';\n'

        return value_format, value

    def write(self, iterable, buff=None):
        buff = self.get_file_obj(buff)

        zip_file = ZipFile(buff, 'w')
        script = StringIO()

        script.write('data SAS_EXPORT;\n')
        script.write('INFILE "data.csv" TRUNCOVER DSD firstobs=2;\n')

        inputs = ''          # field names in sas format
        values = ''          # sas value dictionaries
        value_formats = ''   # labels for value dictionary
        labels = ''          # labels the field names
        informats = ''      # sas informats for all fields
        formats = ''        # sas formats for all fields

        for c in self.concepts:
            cfields = c.concept_fields.select_related('datafield')
            for cfield in cfields:
                field = cfield.field
                name = self._format_name(field.field_name)

                # setting up formats/informats
                sas_form = self._get_formats(name, field)
                formats += sas_form[0]
                informats += sas_form[1]

                # add the field names to the input statement
                inputs += '\t\t' + name
                if field.simple_type == 'string':
                    inputs += ' $'
                inputs += '\n'

                # if a field can be coded create a SAS PROC Format statement
                # that creates a value dictionary
                if field.lexicon:
                    codes = self._code_values(name, field)
                    value_formats += codes[0]
                    values += codes[1]

                # construct labels
                labels += '\tlabel {0}="{1}";\n'.format(name, field.description)

        # Write the SAS File
        script.write(informats + '\n')
        script.write(formats + '\n')
        script.write('input\n' + inputs + ';\n\nrun;\n')
        script.write('proc contents;run;\n\ndata SAS_EXPORT;\n')
        script.write('\tset SAS_EXPORT;\n')
        script.write(labels + '\trun;\n\n')
        script.write('proc format;\n')
        script.write(values + '\nrun;\n\n')
        script.write('data SAS_EXPORT;\n\tset SAS_EXPORT;\n\n')
        script.write(value_formats + 'run;\n\n')
        script.write('/*proc contents data=SAS_EXPORT;*/\n')
        script.write('/*proc print data=SAS_EXPORT;*/\n')
        script.write('run;\nquit;\n')

        zip_file.writestr('export.sas', script.getvalue())
        script.close()

        # Write data CSV
        csv_file = StringIO()
        csv_export = CSVExporter(self.concepts)
        csv_export.preferred_formats = self.preferred_formats
        zip_file.writestr('data.csv', csv_export.write(iterable, csv_file).getvalue())
        csv_file.close()

        zip_file.close()
        return zip_file



from cStringIO import StringIO


class BaseExporter(object):
    "Base class for all exporters"

    preferred_formats = []

    def __init__(self, concepts):
        self.concepts = concepts
        self.params = []

        for concept in concepts:
            length = concept.concept_fields.count()
            self.params.append((concept.format, length))

    def get_file_obj(self, name):
        if name is None:
            return StringIO()
        if isinstance(name, basestring):
            return open(name, 'w+')
        return name

    def _format_row(self, row):
        for formatter, length in self.params:
            values, row = row[:length], row[length:]
            yield formatter(values, self.preferred_formats)

    def read(self, iterable):
        "Takes an iterable that produces rows to be formatted."
        for row in iterable:
            yield self._format_row(row)

    def write(self, iterable, *args, **kwargs):
        raise NotImplemented



from optparse import make_option
from django.core.management.base import BaseCommand
from avocado.core import utils
from avocado.models import DataField, DataConcept, DataCategory
from avocado.legacy import models as legacy


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado legacy [options]

    DESCRIPTION:

        Utilities for migrating Avocado 1.x metadata to the
        Avocado 2.x data model.

    OPTIONS:

        ``--include-non-editable`` - Create ``DataField`` instances for fields marked
        as not editable (i.e. ``editable=False``).
    """

    help = '\n'.join([
        'Utilities for migrating Avocado 1.x metadata to the',
        'Avocado 2.x data model.',
    ])

    option_list = BaseCommand.option_list + (
        make_option('-m', '--migrate', action='store_true',
            dest='migrate', default=False,
            help='Performs the migration'),

        make_option('-e', '--exclude-orphans', action='store_true',
            dest='exclude_orphans', default=False,
            help='During a migration, exclude orphaned fields'),

        make_option('-u', '--update', action='store_true',
            dest='update_existing', default=False,
            help='During a migration, update existing fields'),
    )


    def handle(self, *args, **options):
        migrate = options.get('migrate', False)
        exclude_orphans = options.get('exclude_orphans', False)
        update_existing = options.get('update_existing', False)

        field_cache = {}
        total_migrated = 0

        for lfield in legacy.Field.objects.iterator():
            already_exists = False

            try:
                f = DataField.objects.get_by_natural_key(lfield.app_name,
                    lfield.model_name, lfield.field_name)
                already_exists = True
            except DataField.DoesNotExist:
                f = DataField(app_name=lfield.app_name, model_name=lfield.model_name,
                    field_name=lfield.field_name)

            if already_exists and not update_existing:
                print '({0}) {1}.{2} already exists. Skipping...'.format(f.app_name, f.model_name, f.name)
                continue

            # Map various fields
            f.name = lfield.name,
            f.description = lfield.description
            f.keywords = lfield.keywords
            f.published = lfield.is_public
            f.translator = lfield.translator
            f.group_id = lfield.group_id

            # Check if this is an orphan
            if exclude_orphans and not f.field:
                print '({0}) {1}.{2} is orphaned. Skipping...'.format(f.app_name, f.model_name, f.name)
                continue

            print 'Migrating...\t({0}) {1}.{2}'.format(f.app_name, f.model_name, f.name)

            flags = utils.get_heuristic_flags(f)
            f.__dict__.update(flags)

            # Disagreement with enumerable status
            if f.enumerable != lfield.enable_choices:
                if lfield.enable_choices:
                    override = raw_input('"{0}" is marked as enumerable, but does not qualify to be enumerable. Override? [y/n] ')
                else:
                    override = raw_input('"{0}" is not marked as enumerable, but qualifies to be enumerable. Override? [y/n] ')

                if override.lower() == 'y':
                    f.enumerable = lfield.enable_choices

            if migrate:
                f.save()
                # Cache for Concept processing
                field_cache[f.pk] = f

                # Now that it is saved, associate related objects
                f.sites = lfield.sites.all()
            total_migrated += 1

        print 'Fields migrated:\t{0}'.format(total_migrated)



from django.core.exceptions import ImproperlyConfigured
from avocado.conf import OPTIONAL_DEPS
if not OPTIONAL_DEPS['openpyxl']:
    raise ImproperlyConfigured('openpyxl must be installed to use this exporter.')

from openpyxl import Workbook
from openpyxl.writer.excel import save_virtual_workbook
from _base import BaseExporter


class ExcelExporter(BaseExporter):
    file_extension = 'xlsx'
    preferred_formats = ('boolean', 'number', 'string')

    def write(self, iterable, buff=None):
        buff = self.get_file_obj(buff)

        # Create the workbook and sheets
        wb = Workbook(optimized_write=True)
        ws_data = wb.create_sheet()
        ws_data.title = 'Data'
        ws_dict = wb.create_sheet()
        ws_dict.title = 'Data Dictionary'

        # Create the Data Dictionary Worksheet
        ws_dict.append(('Field Name', 'Data Type', 'Description',
            'Concept Name', 'Concept Discription'))

        for c in self.concepts:
            cfields = c.concept_fields.select_related('field')
            for cfield in cfields:
                datafield = cfield.field
                ws_dict.append((datafield.field_name, datafield.simple_type,
                    datafield.description, c.name, c.description))

        header = []
        # Create the data worksheet
        for i, row_gen in enumerate(self.read(iterable)):
            row = []
            for data in row_gen:
                if i == 0:
                    # Build up header row
                    header.extend(data.keys())
                # Add formatted section to the row
                row.extend(data.values())
            # Write headers on first iteration
            if i == 0:
                ws_data.append(header)
            ws_data.append(row)

        # Save the workbook
        if isinstance(buff, file):
            wb.save(buff)
        else:
            buff.write(save_virtual_workbook(wb))
        return buff



from zipfile import ZipFile
from cStringIO import StringIO
from string import punctuation
from _base import BaseExporter
from _csv import CSVExporter


class RExporter(BaseExporter):
    file_extension = 'zip'
    preferred_formats = ('coded', 'number', 'string')

    def _format_name(self, name):
        punc = punctuation.replace('_', '')
        name = str(name).translate(None, punc)
        name = name.replace(' ', '_')
        words = name.split('_')
        for i, w in enumerate(words):
            if i == 0:
                name = w.lower()
                continue
            name += w.capitalize()
        if name[0].isdigit():
            name = '_' + name

        return name

    def _code_values(self, name, field):
        "If the field can be coded return the r factor and level for it."
        data_field = 'data${0}'.format(name)

        factor = '{0}.factor = factor({0},levels=c('.format(data_field)
        level = 'levels({0}.factor)=c('.format(data_field)

        values_len = len(field.codes)

        for i, (val, code) in enumerate(field.codes):
            factor += str(code)
            level += '"{0}"'.format(str(val))
            if i == values_len - 1:
                factor += '))\n'
                level += ')\n'
                continue
            factor += ' ,'
            level += ' ,'
        return factor, level

    def write(self, iterable, buff=None):
        buff = self.get_file_obj(buff)

        zip_file = ZipFile(buff, 'w')
        script = StringIO()

        script.write('# Read Data\ndata=read.csv("data.csv")\n\n')
        factors = []      # field names
        levels = []       # value dictionaries
        labels = []       # data labels

        for c in self.concepts:
            cfields = c.concept_fields.all()
            for cfield in cfields:
                field = cfield.field
                name = self._format_name(field.field_name)
                labels.append('attr(data${0}, "label") = "{1}"'.format(name, field.description))

                if field.lexicon:
                    codes = self._code_values(name, field)
                    factors.append(codes[0])
                    levels.append(codes[1])

        # Write out the r file to the given scripter
        script.write('# Setting Labels\n')
        script.write('\n'.join(labels) + '\n\n')

        script.write('# Setting Factors\n')
        script.write(''.join(factors) + '\n')

        script.write('# Setting Levels\n')
        script.write(''.join(levels) + '\n')

        zip_file.writestr('export.R', script.getvalue())
        script.close()

        # WRITE CSV
        csv_file = StringIO()
        csv_export = CSVExporter(self.concepts)
        csv_export.preferred_formats = self.preferred_formats
        zip_file.writestr('data.csv', csv_export.write(iterable, csv_file).getvalue())
        csv_file.close()

        zip_file.close()
        return zip_file
